#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import *
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

"""
GRAMPS registration file
"""

MODULE_VERSION = "5.2"

# ------------------------------------------------------------------------
#
# Fix Capitalization of Family Names
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="chname",
    name=_("Fix Capitalization of Family Names"),
    description=_(
        "Searches the entire database and attempts to "
        "fix capitalization of the names."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="changenames.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="ChangeNames",
    optionclass="ChangeNamesOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Rename Event Types
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="chtype",
    name=_("Rename Event Types"),
    description=_(
        "Allows all the events of a certain name " "to be renamed to a new name."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="changetypes.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="ChangeTypes",
    optionclass="ChangeTypesOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Check and Repair Database
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="check",
    name=_("Check and Repair Database"),
    description=_(
        "Checks the database for integrity problems, fixing the " "problems that it can"
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="check.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBFIX,
    toolclass="Check",
    optionclass="CheckOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Compare Individual Events
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="eventcmp",
    name=_("Compare Individual Events"),
    description=_(
        "Aids in the analysis of data by allowing the "
        "development of custom filters that can be applied "
        "to the database to find similar events"
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="eventcmp.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_ANAL,
    toolclass="EventComparison",
    optionclass="EventComparisonOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Extract Event Descriptions from Event Data
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="evname",
    name=_("Extract Event Description"),
    description=_("Extracts event descriptions from the event data"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="eventnames.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="EventNames",
    optionclass="EventNamesOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Find Possible Duplicate People
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="dupfind",
    name=_("Find Possible Duplicate People"),
    description=_(
        "Searches the entire database, looking for "
        "individual entries that may represent the same person."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="finddupes.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="DuplicatePeopleTool",
    optionclass="DuplicatePeopleToolOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Media Manager
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="mediaman",
    name=_("Media Manager"),
    description=_("Manages batch operations on media files"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="mediamanager.py",
    authors=["Alex Roitman"],
    authors_email=["shura@gramps-project.org"],
    category=TOOL_UTILS,
    toolclass="MediaMan",
    optionclass="MediaManOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Not Related
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="not_related",
    name=_("Not Related"),
    description=_(
        "Find people who are not in any way related to the " "selected person"
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="notrelated.py",
    authors=["Stephane Charette"],
    authors_email=["stephanecharette@gmail.com"],
    category=TOOL_UTILS,
    toolclass="NotRelated",
    optionclass="NotRelatedOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Edit Database Owner Information
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="editowner",
    name=_("Edit Database Owner Information"),
    description=_("Allow editing database owner information."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="ownereditor.py",
    authors=["Zsolt Foldvari"],
    authors_email=["zfoldvar@users.sourceforge.net"],
    category=TOOL_DBPROC,
    toolclass="OwnerEditor",
    optionclass="OwnerEditorOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Extract Information from Names
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="patchnames",
    name=_("Extract Information from Names"),
    description=_(
        "Extract titles, prefixes and compound surnames from given name or family name."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="patchnames.py",
    authors=["Donald N. Allingham", "Benny Malengier"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="PatchNames",
    optionclass="PatchNamesOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Rebuild Secondary Indices
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="rebuild",
    name=_("Rebuild Secondary Indexes"),
    description=_("Rebuilds secondary indexes"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="rebuild.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBFIX,
    toolclass="Rebuild",
    optionclass="RebuildOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Rebuild Reference Maps
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="rebuild_refmap",
    name=_("Rebuild Reference Maps"),
    description=_("Rebuilds reference maps"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="rebuildrefmap.py",
    authors=["Alex Roitman"],
    authors_email=["shura@gramps-project.org"],
    category=TOOL_DBFIX,
    toolclass="RebuildRefMap",
    optionclass="RebuildRefMapOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Rebuild Gender Statistics
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="rebuild_genstats",
    name=_("Rebuild Gender Statistics"),
    description=_("Rebuilds gender statistics for name gender guessing..."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="rebuildgenderstat.py",
    authors=["Benny Malengier"],
    authors_email=["benny.malengier@gramps-project.org"],
    category=TOOL_DBFIX,
    toolclass="RebuildGenderStat",
    optionclass="RebuildGenderStatOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Relationship Calculator
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="relcalc",
    name=_("Relationship Calculator"),
    description=_("Calculates the relationship between two people"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="relcalc.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_UTILS,
    toolclass="RelCalc",
    optionclass="RelCalcOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Remove Unused Objects
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="remove_unused",
    name=_("Remove Unused Objects"),
    description=_("Removes unused objects from the database"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="removeunused.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBFIX,
    toolclass="RemoveUnused",
    optionclass="CheckOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Reorder Gramps IDs
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="reorder_ids",
    name=_("Reorder Gramps IDs"),
    description=_("Reorders the Gramps IDs " "according to Gramps' default rules."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="reorderids.py",
    authors=["Donald N. Allingham"],
    authors_email=["don@gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="ReorderIds",
    optionclass="ReorderIdsOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Sort events
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="sortevents",
    name=_("Sort events"),
    description=_("Sort events"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="sortevents.py",
    authors=["Gary Burton"],
    authors_email=["gary.burton@zen.co.uk"],
    category=TOOL_DBPROC,
    toolclass="SortEvents",
    optionclass="SortEventOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Verify the Data
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="verify",
    name=_("Verify the Data"),
    description=_("Verifies the data against user-defined tests"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="verify.py",
    authors=["Alex Roitman"],
    authors_email=["shura@gramps-project.org"],
    category=TOOL_UTILS,
    toolclass="Verify",
    optionclass="VerifyOptions",
    tool_modes=[TOOL_MODE_GUI, TOOL_MODE_CLI],
)

# ------------------------------------------------------------------------
#
# Merge citations
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="mergecitations",
    name=_("Merge Citations"),
    description=_(
        "Searches the entire database, looking for "
        "citations that have the same Volume/Page, Date and Confidence."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="mergecitations.py",
    authors=["Tim G L Lyons"],
    authors_email=["gramps-project.org"],
    category=TOOL_DBPROC,
    toolclass="MergeCitations",
    optionclass="MergeCitationsOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Find database Loop
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="loop",
    name=_("Find database loop"),
    description=_("Searches the entire database, looking for " "a possible loop."),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="findloop.py",
    authors=["Serge Noiraud"],
    authors_email=["serge.noiraud@free.fr"],
    category=TOOL_UTILS,
    toolclass="FindLoop",
    optionclass="FindLoopOptions",
    tool_modes=[TOOL_MODE_GUI],
)

# ------------------------------------------------------------------------
#
# Remove leading and trailing spaces for places name
# Remove leading and trailing spaces for surname and first names
#
# ------------------------------------------------------------------------

register(
    TOOL,
    id="removespaces",
    name=_("Clean input data"),
    description=_(
        "Searches the entire database, looking for "
        "trailing or leading spaces for places and people."
        " Search comma in coordinates fields in places."
    ),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="removespaces.py",
    authors=["Serge Noiraud"],
    authors_email=["serge.noiraud@free.fr"],
    category=TOOL_UTILS,
    toolclass="RemoveSpaces",
    optionclass="RemoveSpacesOptions",
    tool_modes=[TOOL_MODE_GUI],
)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011 Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.editors import EditSource, EditCitation
from gramps.gui.listmodel import ListModel, NOSORT
from gramps.gen.plug import Gramplet
from gramps.gen.utils.string import conf_strings
from gramps.gen.datehandler._dateutils import get_date
from gramps.gui.dbguielement import DbGUIElement
from gramps.gen.errors import WindowActiveError
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gui.widgets.persistenttreeview import PersistentTreeView

_ = glocale.translation.gettext


class Citations(Gramplet, DbGUIElement):
    """
    Displays the citations for an object.
    """

    def __init__(self, gui, nav_group=0):
        Gramplet.__init__(self, gui, nav_group)
        DbGUIElement.__init__(self, self.dbstate.db)
        self.source_nodes = {}

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()
        self.gui.WIDGET.restore_column_size()

    def on_save(self):
        self.gui.WIDGET.save_column_info()

    def _connect_db_signals(self):
        """
        called on init of DbGUIElement, connect to db as required.
        """
        self.callman.register_callbacks(
            {
                "citation-update": self.changed,
                "person-update": self.changed,
                "family-update": self.changed,
                "event-update": self.changed,
                "media-update": self.changed,
                "place-update": self.changed,
            }
        )
        self.callman.connect_all(
            keys=["citation", "family", "person", "event", "media", "place"]
        )

    def changed(self, handle):
        """
        Called when a registered citation is updated.
        """
        self.update()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        tip = _("Double-click on a row to edit the selected source/citation.")
        self.set_tooltip(tip)
        top = PersistentTreeView(self.uistate, __name__)
        titles = [
            (
                "",
                NOSORT,
                50,
            ),
            (_("Source/Date"), 1, 350),
            (_("Volume/Page"), 2, 150),
            (_("Confidence Level"), 3, 150),
            (_("Author"), 4, 200),
            (_("Publisher"), 5, 150),
        ]
        self.model = ListModel(
            top, titles, list_mode="tree", event_func=self.invoke_editor
        )
        return top

    def add_citations(self, obj):
        for citation_handle in obj.get_citation_list():
            self.add_citation_ref(citation_handle)

    def add_name_citations(self, obj):
        names = [obj.get_primary_name()] + obj.get_alternate_names()
        for name in names:
            self.add_citations(name)

    def add_attribute_citations(self, obj):
        for attr in obj.get_attribute_list():
            self.add_citations(attr)

    def add_mediaref_citations(self, obj):
        for media_ref in obj.get_media_list():
            self.add_citations(media_ref)
            self.add_attribute_citations(media_ref)
            media = self.dbstate.db.get_media_from_handle(media_ref.ref)
            self.add_media_citations(media)

    def add_media_citations(self, media):
        self.callman.register_handles({"media": [media.handle]})
        self.add_citations(media)
        self.add_attribute_citations(media)

    def add_eventref_citations(self, obj):
        for event_ref in obj.get_event_ref_list():
            self.add_citations(event_ref)
            self.add_attribute_citations(event_ref)
            event = self.dbstate.db.get_event_from_handle(event_ref.ref)
            self.add_event_citations(event)

    def add_event_citations(self, event):
        self.callman.register_handles({"event": [event.handle]})
        self.add_citations(event)
        self.add_attribute_citations(event)
        self.add_mediaref_citations(event)
        place_handle = event.get_place_handle()
        if place_handle:
            place = self.dbstate.db.get_place_from_handle(place_handle)
            if place:
                self.add_place_citations(place)

    def add_place_citations(self, place):
        self.callman.register_handles({"place": [place.handle]})
        self.add_citations(place)
        self.add_mediaref_citations(place)

    def add_address_citations(self, obj):
        for address in obj.get_address_list():
            self.add_citations(address)

    def add_lds_citations(self, obj):
        for lds in obj.get_lds_ord_list():
            self.add_citations(lds)
            place_handle = lds.get_place_handle()
            if place_handle:
                place = self.dbstate.db.get_place_from_handle(place_handle)
                if place:
                    self.add_place_citations(place)

    def add_association_citations(self, obj):
        for assoc in obj.get_person_ref_list():
            self.add_citations(assoc)

    def add_citation_ref(self, citation_handle):
        """
        Add a citation to the model.
        """
        self.callman.register_handles({"citation": [citation_handle]})
        citation = self.dbstate.db.get_citation_from_handle(citation_handle)
        page = citation.get_page()
        if not page:
            page = _("<No Volume/Page>")
        source_handle = citation.get_reference_handle()
        source = self.dbstate.db.get_source_from_handle(source_handle)
        title = source.get_title()
        author = source.get_author()
        publisher = source.get_publication_info()
        confidence = citation.get_confidence_level()

        if source_handle not in self.source_nodes:
            node = self.model.add([source_handle, title, "", "", author, publisher])
            self.source_nodes[source_handle] = node

        self.model.add(
            [
                citation_handle,
                get_date(citation),
                page,
                _(conf_strings[confidence]),
                "",
                "",
            ],
            node=self.source_nodes[source_handle],
        )

    def check_citations(self, obj):
        return True if obj.get_citation_list() else False

    def check_name_citations(self, obj):
        names = [obj.get_primary_name()] + obj.get_alternate_names()
        for name in names:
            if self.check_citations(name):
                return True
        return False

    def check_attribute_citations(self, obj):
        for attr in obj.get_attribute_list():
            if self.check_citations(attr):
                return True
        return False

    def check_mediaref_citations(self, obj):
        for media_ref in obj.get_media_list():
            if self.check_citations(media_ref):
                return True
            if self.check_attribute_citations(media_ref):
                return True
            media = self.dbstate.db.get_media_from_handle(media_ref.ref)
            if self.check_media_citations(media):
                return True
        return False

    def check_media_citations(self, media):
        if self.check_citations(media):
            return True
        if self.check_attribute_citations(media):
            return True
        return False

    def check_eventref_citations(self, obj):
        if obj:
            for event_ref in obj.get_event_ref_list():
                if self.check_attribute_citations(event_ref):
                    return True
                event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                if self.check_event_citations(event):
                    return True
        return False

    def check_event_citations(self, event):
        if self.check_citations(event):
            return True
        if self.check_attribute_citations(event):
            return True
        if self.check_mediaref_citations(event):
            return True
        place_handle = event.get_place_handle()
        if place_handle:
            place = self.dbstate.db.get_place_from_handle(place_handle)
            if place and self.check_place_citations(place):
                return True
        return False

    def check_place_citations(self, place):
        if self.check_citations(place):
            return True
        if self.check_mediaref_citations(place):
            return True
        return False

    def check_address_citations(self, obj):
        for address in obj.get_address_list():
            if self.check_citations(address):
                return True
        return False

    def check_lds_citations(self, obj):
        for lds in obj.get_lds_ord_list():
            if self.check_citations(lds):
                return True
            place_handle = lds.get_place_handle()
            if place_handle:
                place = self.dbstate.db.get_place_from_handle(place_handle)
                if place and self.check_place_citations(place):
                    return True
        return False

    def check_association_citations(self, obj):
        for assoc in obj.get_person_ref_list():
            if self.check_citations(assoc):
                return True
        return False

    def invoke_editor(self, treeview):
        """
        Edit the selected source or citation.
        """
        model, iter_ = treeview.get_selection().get_selected()
        if iter_:
            handle = model.get_value(iter_, 0)
            # bug 9094.
            # str(model.get_path(iter_)) return something like NNN:MMM
            # So if we have only NNN, it's a node
            # removing the str() solves the problem.
            if len(model.get_path(iter_)) == 1:
                self.edit_source(handle)
            else:
                self.edit_citation(handle)

    def edit_source(self, handle):
        """
        Edit the selected source.
        """
        try:
            source = self.dbstate.db.get_source_from_handle(handle)
            EditSource(self.dbstate, self.uistate, [], source)
        except WindowActiveError:
            pass

    def edit_citation(self, handle):
        """
        Edit the selected citation.
        """
        try:
            citation = self.dbstate.db.get_citation_from_handle(handle)
            source_handle = citation.get_reference_handle()
            source = self.dbstate.db.get_source_from_handle(source_handle)
            EditCitation(self.dbstate, self.uistate, [], citation, source)
        except WindowActiveError:
            pass


class PersonCitations(Citations):
    """
    Displays the citations for a person.
    """

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        self.callman.unregister_all()
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            if active:
                self.display_citations(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_citations(self, person):
        """
        Display the citations for the active person.
        """
        self.source_nodes = {}
        self.callman.register_handles({"person": [person.handle]})
        self.add_citations(person)
        self.add_eventref_citations(person)
        for handle in person.get_family_handle_list():
            family = self.dbstate.db.get_family_from_handle(handle)
            self.add_eventref_citations(family)
        self.add_name_citations(person)
        self.add_attribute_citations(person)
        self.add_address_citations(person)
        self.add_mediaref_citations(person)
        self.add_association_citations(person)
        self.add_lds_citations(person)

        self.set_has_data(self.model.count > 0)
        self.model.tree.expand_all()

    def get_has_data(self, person):
        """
        Return True if the gramplet has data, else return False.
        """
        if person is None:
            return False
        if self.check_citations(person):
            return True
        if self.check_eventref_citations(person):
            return True
        for handle in person.get_family_handle_list():
            family = self.dbstate.db.get_family_from_handle(handle)
            if self.check_eventref_citations(family):
                return True
        if self.check_name_citations(person):
            return True
        if self.check_attribute_citations(person):
            return True
        if self.check_address_citations(person):
            return True
        if self.check_mediaref_citations(person):
            return True
        if self.check_association_citations(person):
            return True
        if self.check_lds_citations(person):
            return True
        return False


class EventCitations(Citations):
    """
    Displays the citations for an event.
    """

    def db_changed(self):
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        self.callman.unregister_all()
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            if active:
                self.display_citations(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_citations(self, event):
        """
        Display the citations for the active event.
        """
        self.source_nodes = {}
        self.add_event_citations(event)
        self.set_has_data(self.model.count > 0)
        self.model.tree.expand_all()

    def get_has_data(self, event):
        """
        Return True if the gramplet has data, else return False.
        """
        if event is None:
            return False
        if self.check_event_citations(event):
            return True
        return False


class FamilyCitations(Citations):
    """
    Displays the citations for a family.
    """

    def db_changed(self):
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        self.callman.unregister_all()
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            if active:
                self.display_citations(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_citations(self, family):
        """
        Display the citations for the active family.
        """
        self.source_nodes = {}
        self.callman.register_handles({"family": [family.handle]})
        self.add_citations(family)
        self.add_eventref_citations(family)
        self.add_attribute_citations(family)
        self.add_mediaref_citations(family)
        self.add_lds_citations(family)

        self.set_has_data(self.model.count > 0)
        self.model.tree.expand_all()

    def get_has_data(self, family):
        """
        Return True if the gramplet has data, else return False.
        """
        if family is None:
            return False
        if self.check_citations(family):
            return True
        if self.check_eventref_citations(family):
            return True
        if self.check_attribute_citations(family):
            return True
        if self.check_mediaref_citations(family):
            return True
        if self.check_lds_citations(family):
            return True
        return False


class PlaceCitations(Citations):
    """
    Displays the citations for a place.
    """

    def db_changed(self):
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        self.callman.unregister_all()
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            if active:
                self.display_citations(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_citations(self, place):
        """
        Display the citations for the active place.
        """
        self.source_nodes = {}
        self.add_place_citations(place)
        self.set_has_data(self.model.count > 0)
        self.model.tree.expand_all()

    def get_has_data(self, place):
        """
        Return True if the gramplet has data, else return False.
        """
        if place is None:
            return False
        if self.check_place_citations(place):
            return True
        return False


class MediaCitations(Citations):
    """
    Displays the citations for a media object.
    """

    def db_changed(self):
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.model.clear()
        self.callman.unregister_all()
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            if active:
                self.display_citations(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def display_citations(self, media):
        """
        Display the citations for the active media object.
        """
        self.source_nodes = {}
        self.add_media_citations(media)
        self.set_has_data(self.model.count > 0)
        self.model.tree.expand_all()

    def get_has_data(self, media):
        """
        Return True if the gramplet has data, else return False.
        """
        if media is None:
            return False
        if self.check_media_citations(media):
            return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2007-2009  Brian G. Matherly
# Copyright (C) 2009-2010  Benny Malengier <benny.malengier@gramps-project.org>
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2012,2017  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
ACSII document generator.
"""

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import DOCGEN_OPTIONS
from gramps.gen.errors import ReportError
from gramps.gen.plug.docgen import BaseDoc, TextDoc, PARA_ALIGN_RIGHT, PARA_ALIGN_CENTER
from gramps.gen.plug.menu import NumberOption
from gramps.gen.plug.report import DocOptions
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------
LEFT, RIGHT, CENTER = "LEFT", "RIGHT", "CENTER"


# ------------------------------------------------------------------------
#
# This routine was written by David Mertz and placed into the public
# domain. It is sample code from his book, "Text Processing in Python"
#
# Modified by Alex Roitman: right-pad with spaces, if right_pad==1;
#                           return empty string if no text was given
# Another argument: "first" is the first line indent in characters
#                   _relative_ to the "left" margin. It can be negative!
#
# ------------------------------------------------------------------------
def reformat_para(para="", left=0, right=72, just=LEFT, right_pad=0, first=0):
    if not para.strip():
        return "\n"

    lines = []
    real_left = left + first
    alllines = para.split("\n")
    for realline in alllines:
        words = realline.split()
        line = ""
        word = 0
        end_words = 0
        while not end_words:
            if not words:
                lines.append("\n")
                break
            if len(words[word]) > right - real_left:  # Handle very long words
                line = words[word]
                word += 1
                if word >= len(words):
                    end_words = 1
            else:  # Compose line of words
                while len(line) + len(words[word]) <= right - real_left:
                    line += words[word]
                    word += 1
                    if word >= len(words):
                        end_words = 1
                        break
                    elif len(line) < right - real_left:
                        line += " "  # add a space since there is still room
            lines.append(line)
            # first line finished, discard first
            real_left = left
            line = ""
    if just == CENTER:
        if right_pad:
            return "\n".join(
                [
                    " " * (left + first) + ln.center(right - left - first)
                    for ln in lines[0:1]
                ]
                + [" " * left + ln.center(right - left) for ln in lines[1:]]
            )
        else:
            return "\n".join(
                [
                    " " * (left + first) + ln.center(right - left - first).rstrip()
                    for ln in lines[0:1]
                ]
                + [" " * left + ln.center(right - left).rstrip() for ln in lines[1:]]
            )
    elif just == RIGHT:
        if right_pad:
            return "\n".join([line.rjust(right) for line in lines])
        else:
            return "\n".join([line.rjust(right).rstrip() for line in lines])
    else:  # left justify
        if right_pad:
            return "\n".join(
                [
                    " " * (left + first) + line.ljust(right - left - first)
                    for line in lines[0:1]
                ]
                + [" " * left + line.ljust(right - left) for line in lines[1:]]
            )
        else:
            return "\n".join(
                [" " * (left + first) + line for line in lines[0:1]]
                + [" " * left + line for line in lines[1:]]
            )


# ------------------------------------------------------------------------
#
# Ascii
#
# ------------------------------------------------------------------------
class AsciiDoc(BaseDoc, TextDoc):
    """
    ASCII document generator.
    """

    def __init__(self, styles, paper_style, options=None, uistate=None):
        BaseDoc.__init__(self, styles, paper_style, uistate=uistate)
        self.__note_format = False

        self._cpl = 72  # characters per line, in case the options are ignored
        if options:
            menu = options.menu
            self._cpl = menu.get_option_by_name("linechars").get_value()

        self.file = None
        self.filename = ""

        self.text = ""
        self.para = None
        self.leader = None

        self.tbl_style = None
        self.in_cell = None
        self.ncols = 0
        self.column_order = []
        self.cellpars = []
        self.cell_lines = []
        self.cell_widths = []
        self.cellnum = -1
        self.maxlines = 0

    # --------------------------------------------------------------------
    #
    # Opens the file, resets the text buffer.
    #
    # --------------------------------------------------------------------
    def open(self, filename):
        if filename[-4:] != ".txt":
            self.filename = filename + ".txt"
        else:
            self.filename = filename

        try:
            self.file = open(self.filename, "w", errors="backslashreplace")
        except Exception as msg:
            raise ReportError(_("Could not create %s") % self.filename, msg)

        self.in_cell = 0
        self.text = ""

    # --------------------------------------------------------------------
    #
    # Close the file. Call the app if required.
    #
    # --------------------------------------------------------------------
    def close(self):
        self.file.close()

    def get_usable_width(self):
        """
        Return the usable width of the document in characters.
        """
        return self._cpl

    # --------------------------------------------------------------------
    #
    # Force a section page break
    #
    # --------------------------------------------------------------------
    def page_break(self):
        self.file.write("\012")

    def start_bold(self):
        pass

    def end_bold(self):
        pass

    def start_superscript(self):
        self.text = self.text + "["

    def end_superscript(self):
        self.text = self.text + "]"

    # --------------------------------------------------------------------
    #
    # Starts a paragraph.
    #
    # --------------------------------------------------------------------
    def start_paragraph(self, style_name, leader=None):
        styles = self.get_style_sheet()
        self.para = styles.get_paragraph_style(style_name)
        self.leader = leader

    # --------------------------------------------------------------------
    #
    # End a paragraph. First format it to the desired widths.
    # If not in table cell, write it immediately. If in the cell,
    # add it to the list for this cell after formatting.
    #
    # --------------------------------------------------------------------
    def end_paragraph(self):
        if self.para.get_alignment() == PARA_ALIGN_RIGHT:
            fmt = RIGHT
        elif self.para.get_alignment() == PARA_ALIGN_CENTER:
            fmt = CENTER
        else:
            fmt = LEFT

        if self.in_cell:
            right = self.cell_widths[self.cellnum]
        else:
            right = self.get_usable_width()

        # Compute indents in characters. Keep first_indent relative!
        regular_indent = 0
        first_indent = 0
        if self.para.get_left_margin():
            regular_indent = int(4 * self.para.get_left_margin())
        if self.para.get_first_indent():
            first_indent = int(4 * self.para.get_first_indent())

        if self.in_cell and self.cellnum < self.ncols - 1:
            right_pad = 1
            the_pad = " " * right
        else:
            right_pad = 0
            the_pad = ""

        # Depending on the leader's presence, treat the first line differently
        if self.leader:
            # If we have a leader then we need to reformat the text
            # as if there's no special treatment for the first line.
            # Then add leader and eat up the beginning of the first line pad.
            # Do not reformat if preformatted notes
            if not self.__note_format:
                self.leader += " "
                start_at = regular_indent + min(len(self.leader) + first_indent, 0)
                this_text = reformat_para(
                    self.text, regular_indent, right, fmt, right_pad
                )
                this_text = (
                    " " * (regular_indent + first_indent)
                    + self.leader
                    + this_text[start_at:]
                )
            else:
                this_text = self.text
        else:
            # If no leader then reformat the text according to the first
            # line indent, as specified by style.
            # Do not reformat if preformatted notes
            if not self.__note_format:
                this_text = reformat_para(
                    self.text, regular_indent, right, fmt, right_pad, first_indent
                )
            else:
                this_text = " " * (regular_indent + first_indent) + self.text

        if self.__note_format:
            # don't add an extra LF before the_pad if preformatted notes.
            if this_text != "\n":
                # don't add LF if there is this_text is a LF
                this_text += the_pad + "\n"
        else:
            this_text += "\n" + the_pad + "\n"

        if self.in_cell:
            self.cellpars[self.cellnum] += this_text
        else:
            self.file.write(this_text)

        self.text = ""

    # --------------------------------------------------------------------
    #
    # Start a table. Grab the table style, and store it.
    #
    # --------------------------------------------------------------------
    def start_table(self, name, style_name):
        styles = self.get_style_sheet()
        self.tbl_style = styles.get_table_style(style_name)
        self.ncols = self.tbl_style.get_columns()
        self.column_order = []
        for cell in range(self.ncols):
            self.column_order.append(cell)
        if self.get_rtl_doc():
            self.column_order.reverse()

    # --------------------------------------------------------------------
    #
    # End a table. Turn off the self.in_cell flag
    #
    # --------------------------------------------------------------------
    def end_table(self):
        self.in_cell = 0

    # --------------------------------------------------------------------
    #
    # Start a row. Initialize lists for cell contents, number of lines,
    # and the widths. It is necessary to keep a list of cell contents
    # that is to be written after all the cells are defined.
    #
    # --------------------------------------------------------------------
    def start_row(self):
        self.cellpars = [""] * self.ncols
        self.cell_lines = [0] * self.ncols
        self.cell_widths = [0] * self.ncols
        self.cellnum = -1
        self.maxlines = 0
        table_width = self.get_usable_width() * self.tbl_style.get_width() / 100.0
        for cell in self.column_order:
            self.cell_widths[cell] = int(
                table_width * self.tbl_style.get_column_width(cell) / 100.0
            )

    # --------------------------------------------------------------------
    #
    # End a row. Write the cell contents. Write the line of spaces
    # if the cell has fewer lines than the maximum number.
    #
    # --------------------------------------------------------------------
    def end_row(self):
        self.in_cell = 0
        cell_text = [None] * self.ncols
        for cell in self.column_order:
            if self.cell_widths[cell]:
                blanks = " " * self.cell_widths[cell] + "\n"
                if self.cell_lines[cell] < self.maxlines:
                    self.cellpars[cell] += blanks * (
                        self.maxlines - self.cell_lines[cell]
                    )
                cell_text[cell] = self.cellpars[cell].split("\n")
        for line in range(self.maxlines):
            for cell in self.column_order:
                if self.cell_widths[cell]:
                    self.file.write(cell_text[cell][line])
            self.file.write("\n")

    # --------------------------------------------------------------------
    #
    # Start a cell. Set the self.in_cell flag,
    # increment the current cell number.
    #
    # --------------------------------------------------------------------
    def start_cell(self, style_name, span=1):
        self.in_cell = 1
        self.cellnum = self.cellnum + span
        span -= 1
        while span:
            self.cell_widths[self.cellnum] += self.cell_widths[self.cellnum - span]
            self.cell_widths[self.cellnum - span] = 0
            span -= 1

    # --------------------------------------------------------------------
    #
    # End a cell. Find out the number of lines in this cell, correct
    # the maximum number of lines if necessary.
    #
    # --------------------------------------------------------------------
    def end_cell(self):
        self.in_cell = 0
        self.cell_lines[self.cellnum] = self.cellpars[self.cellnum].count("\n")
        if self.cell_lines[self.cellnum] > self.maxlines:
            self.maxlines = self.cell_lines[self.cellnum]

    def add_media(self, name, align, w_cm, h_cm, alt="", style_name=None, crop=None):
        this_text = "(photo)"
        if self.in_cell:
            self.cellpars[self.cellnum] += this_text
        else:
            self.file.write(this_text)

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the ASCII doc.
        styledtext : assumed a StyledText object to write
        format : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        contains_html: bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable. AsciiDoc prints the html without handling it
        links: bool, make the URL in the text clickable (if supported)
        """
        if contains_html:
            return
        text = str(styledtext)
        if format:
            # Preformatted note, keep all white spaces, tabs, LF's
            self.__note_format = True
            for line in text.split("\n"):
                self.start_paragraph(style_name)
                self.write_text(line)
                self.end_paragraph()
            # Add an extra empty para all lines in each preformatted note
            self.start_paragraph(style_name)
            self.end_paragraph()
            self.__note_format = False
        else:
            for line in text.split("\n\n"):
                self.start_paragraph(style_name)
                # line = line.replace('\n',' ')
                # line = ' '.join(line.split())
                self.write_text(line)
                self.end_paragraph()

    # --------------------------------------------------------------------
    #
    # Writes text.
    # --------------------------------------------------------------------
    def write_text(self, text, mark=None, links=False):
        self.text = self.text + text


# ------------------------------------------------------------------------
#
# AsciiDocOptions class
#
# ------------------------------------------------------------------------
class AsciiDocOptions(DocOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        DocOptions.__init__(self, name)

    def add_menu_options(self, menu):
        """
        Add options to the document menu for the AsciiDoc docgen.
        """

        category_name = DOCGEN_OPTIONS

        linechars = NumberOption(_("Characters per line"), 72, 20, 9999)
        linechars.set_help(_("The number of characters per line"))
        menu.add_option(category_name, "linechars", linechars)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2002       Gary Shao
# Copyright (C) 2007       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".textdoc")

# -------------------------------------------------------------------------
#
# URL string pattern
#
# -------------------------------------------------------------------------
URL_PATTERN = (
    r"""(((https?|mailto):)(//([^ /?#"]*))?([^ ?#"]*)(\?([^ #"]*))?(#([^ "]*))?)"""
)

# -------------------------------------------------------------------------
#
# IndexMark types
#
# -------------------------------------------------------------------------
INDEX_TYPE_ALP = 0
INDEX_TYPE_TOC = 1
LOCAL_HYPERLINK = 2
LOCAL_TARGET = 3


# ------------------------------------------------------------------------
#
# IndexMark
#
# ------------------------------------------------------------------------
class IndexMark:
    """
    Defines a mark to be associated with text for indexing.
    """

    def __init__(self, key="", itype=INDEX_TYPE_ALP, level=1):
        """
        Initialize the object with default values, unless values are specified.
        """
        self.key = key
        self.type = itype
        self.level = level


# ------------------------------------------------------------------------
#
# TextDoc
#
# ------------------------------------------------------------------------


class TextDoc(metaclass=ABCMeta):
    """
    Abstract Interface for text document generators. Output formats for
    text reports must implement this interface to be used by the report
    system.
    """

    @abstractmethod
    def page_break(self):
        """
        Forces a page break, creating a new page.
        """

    @abstractmethod
    def start_bold(self):
        """
        Start a section of bold text.
        """

    @abstractmethod
    def end_bold(self):
        """
        End a section of bold text.
        """

    @abstractmethod
    def start_superscript(self):
        """
        Start a section of superscript text.
        """

    @abstractmethod
    def end_superscript(self):
        """
        End a section of superscript text.
        """

    @abstractmethod
    def start_paragraph(self, style_name, leader=None):
        """
        Starts a new paragraph, using the specified style name.

        :param style_name: name of the :class:`.ParagraphStyle` to use for the
                           paragraph.
        :param leader: Leading text for a paragraph. Typically used
                       for numbering.
        """

    @abstractmethod
    def end_paragraph(self):
        """
        Ends the current paragraph.
        """

    @abstractmethod
    def start_table(self, name, style_name):
        """
        Starts a new table.

        :param name: Unique name of the table.
        :param style_name: :class:`.TableStyle` to use for the new table
        """

    @abstractmethod
    def end_table(self):
        """
        Ends the current table.
        """

    @abstractmethod
    def start_row(self):
        """
        Starts a new row on the current table.
        """

    @abstractmethod
    def end_row(self):
        """
        Ends the current row on the current table.
        """

    @abstractmethod
    def start_cell(self, style_name, span=1):
        """
        Starts a new table cell, using the paragraph style specified.

        :param style_name: :class:`.TableCellStyle` to use for the cell
        :param span: number of columns to span
        """

    @abstractmethod
    def end_cell(self):
        """
        Ends the current table cell.
        """

    @abstractmethod
    def write_text(self, text, mark=None, links=False):
        """
        Writes the text in the current paragraph. Should only be used after a
        start_paragraph and before an end_paragraph.

        :param text: text to write.
        :param mark:  :class:`.IndexMark` to use for indexing (if supported)
        :param links: make URLs in the text clickable (if supported)
        """

    def write_markup(self, text, s_tags, mark=None):
        """
        Writes the text in the current paragraph.  Should only be used after a
        start_paragraph and before an end_paragraph. Not all backends support
        s_tags, then the same happens as with write_text. Backends supporting
        write_markup will overwrite this method

        :param text: text to write. The text is assumed to be _not_ escaped
        :param s_tags: assumed to be list of styledtexttags to apply to the
                       text
        :param mark:  :class:`.IndexMark` to use for indexing (if supported)
        """
        self.write_text(text, mark=mark)

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the cairo doc.

        :param styledtext: assumed a :class:`.StyledText` object to write
        :param format: 0 = Flowed, 1 = Preformatted
        :param style_name: name of the style to use for default presentation
        :param contains_html:
            bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable.
        :param links: bool, make URLs in the text clickable (if supported)

        overwrite this method if the backend supports styled notes
        """
        text = str(styledtext)
        self.write_note(text, format, style_name)

    def write_text_citation(self, text, mark=None, links=None):
        """
        Method to write text with Gramps <super> citation marks.
        """
        if not text:
            return
        parts = text.split("<super>")
        markset = False
        for piece in parts:
            if not piece:
                # a text '<super>text ...' splits as '', 'text..'
                continue
            piecesplit = piece.split("</super>")
            if len(piecesplit) == 2:
                self.start_superscript()
                self.write_text(piecesplit[0], links=links)
                self.end_superscript()
                if not piecesplit[1]:
                    # text ended with ' ... </super>'
                    continue
                if not markset:
                    self.write_text(piecesplit[1], mark, links=links)
                    markset = True
                else:
                    self.write_text(piecesplit[1], links=links)
            else:
                if not markset:
                    self.write_text(piece, mark, links=links)
                    markset = True
                else:
                    self.write_text(piece, links=links)

    @abstractmethod
    def add_media(self, name, align, w_cm, h_cm, alt="", style_name=None, crop=None):
        """
        Add a photo of the specified width (in centimeters).

        :param name: filename of the image to add
        :param align: alignment of the image. Valid values are 'left',
                      'right', 'center', and 'single'
        :param w_cm: width in centimeters
        :param h_cm: height in centimeters
        :param alt: an alternative text to use. Useful for eg html reports
        :param style_name: style to use for captions
        :param crop: image cropping parameters
        """

    def start_link(self, link):
        """
        Start a link section. This defaults to underlining.

        :param link: should be an item that makes sense in this
                     docgen type, if it implements linking.
        """
        self.start_underline()

    def stop_link(self):
        """
        Stop the link section. Defaults to stopping the underlining
        for docgen types that don't support links.
        """
        self.stop_underline()

    def start_underline(self):
        """
        Start a section of underlining. This passes without error
        so that docgen types are not required to have this.
        """
        pass

    def stop_underline(self):
        """
        Stops a section of underlining. This passes without error
        so that docgen ntypes are not required to have this.
        """
        pass

    def insert_toc(self):
        """
        Insert a Table of Contents at this point in the document. This passes
        without error so that docgen types are not required to have this.
        """
        pass

    def insert_index(self):
        """
        Insert an Alphabetical Index at this point in the document. This passes
        without error so that docgen types are not required to have this.
        """
        pass



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2012       Doug Blank
# Copyright (C) 2013       John Ralls <jralls@ceridwen.us>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provides constants for other modules
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import sys
import uuid

from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .git_revision import get_git_revision
from .constfunc import get_env_var
from ..version import VERSION, VERSION_TUPLE, major_version, DEV_VERSION
from .utils.resourcepath import ResourcePath
from .utils.grampslocale import GrampsLocale

# -------------------------------------------------------------------------
#
# Gramps program name
#
# -------------------------------------------------------------------------
PROGRAM_NAME = "Gramps"

# -------------------------------------------------------------------------
#
# Standard Gramps Websites
#
# -------------------------------------------------------------------------
URL_HOMEPAGE = "http://gramps-project.org/"
URL_MAILINGLIST = "http://sourceforge.net/mail/?group_id=25770"
URL_BUGHOME = "http://gramps-project.org/bugs"
URL_BUGTRACKER = "http://gramps-project.org/bugs/bug_report_page.php"
URL_WIKISTRING = "http://gramps-project.org/wiki/index.php?title="
URL_MANUAL_PAGE = "Gramps_%s_Wiki_Manual" % major_version
URL_MANUAL_DATA = "%s_-_Entering_and_editing_data:_detailed" % URL_MANUAL_PAGE
URL_MANUAL_SECT1 = "%s_-_part_1" % URL_MANUAL_DATA
URL_MANUAL_SECT2 = "%s_-_part_2" % URL_MANUAL_DATA
URL_MANUAL_SECT3 = "%s_-_part_3" % URL_MANUAL_DATA
WIKI_FAQ = "FAQ"
WIKI_KEYBINDINGS = "Gramps_%s_Wiki_Manual_-_Keybindings" % major_version
WIKI_EXTRAPLUGINS = "%s_Addons" % major_version
WIKI_EXTRAPLUGINS_RAWDATA = "Plugins%s&action=raw" % major_version

# -------------------------------------------------------------------------
#
# Mime Types
#
# -------------------------------------------------------------------------
APP_FAMTREE = "x-directory/normal"
APP_GRAMPS = "application/x-gramps"
APP_GRAMPS_XML = "application/x-gramps-xml"
APP_GEDCOM = "application/x-gedcom"
APP_GRAMPS_PKG = "application/x-gramps-package"
APP_GENEWEB = "application/x-geneweb"
APP_VCARD = ["text/x-vcard", "text/x-vcalendar"]

# -------------------------------------------------------------------------
#
# Determine the home directory. According to Wikipedia, most UNIX like
# systems use HOME. I'm assuming that this would apply to OS X as well.
# Windows apparently uses USERPROFILE
#
# -------------------------------------------------------------------------
if "GRAMPSHOME" in os.environ:
    USER_HOME = get_env_var("GRAMPSHOME")
    HOME_DIR = os.path.join(USER_HOME, "gramps")
elif "USERPROFILE" in os.environ:
    USER_HOME = get_env_var("USERPROFILE")
    if "APPDATA" in os.environ:
        HOME_DIR = os.path.join(get_env_var("APPDATA"), "gramps")
    else:
        HOME_DIR = os.path.join(USER_HOME, "gramps")
else:
    USER_HOME = get_env_var("HOME")
    HOME_DIR = os.path.join(USER_HOME, ".gramps")
ORIG_HOME_DIR = HOME_DIR
if "SAFEMODE" in os.environ:
    if "USERPROFILE" in os.environ:
        USER_HOME = get_env_var("USERPROFILE")
    else:
        USER_HOME = get_env_var("HOME")
    HOME_DIR = get_env_var("SAFEMODE")


if os.path.exists(HOME_DIR) or "GRAMPSHOME" in os.environ or "SAFEMODE" in os.environ:
    USER_DATA = HOME_DIR
    USER_CONFIG = HOME_DIR
    USER_CACHE = HOME_DIR
else:
    USER_DATA = os.path.join(GLib.get_user_data_dir(), "gramps")
    USER_CONFIG = os.path.join(GLib.get_user_config_dir(), "gramps")
    USER_CACHE = os.path.join(GLib.get_user_cache_dir(), "gramps")

USER_PICTURES = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES)
if not USER_PICTURES:
    USER_PICTURES = HOME_DIR

VERSION_DIR_NAME = "gramps%s%s" % (VERSION_TUPLE[0], VERSION_TUPLE[1])
VERSION_DIR = os.path.join(USER_CONFIG, VERSION_DIR_NAME)
USER_DATA_VERSION = os.path.join(USER_DATA, VERSION_DIR_NAME)

CUSTOM_FILTERS = os.path.join(VERSION_DIR, "custom_filters.xml")
REPORT_OPTIONS = os.path.join(USER_CONFIG, "report_options.xml")
TOOL_OPTIONS = os.path.join(USER_CONFIG, "tool_options.xml")
PLACE_FORMATS = os.path.join(USER_CONFIG, "place_formats.xml")

THUMB_DIR = os.path.join(USER_CACHE, "thumb")
THUMB_NORMAL = os.path.join(THUMB_DIR, "normal")
THUMB_LARGE = os.path.join(THUMB_DIR, "large")
USER_PLUGINS = os.path.join(USER_DATA_VERSION, "plugins")
USER_CSS = os.path.join(USER_DATA, "css")
# dirs checked/made for each Gramps session
USER_DIRLIST = (
    USER_HOME,
    USER_CACHE,
    USER_CONFIG,
    USER_DATA,
    VERSION_DIR,
    USER_DATA_VERSION,
    THUMB_DIR,
    THUMB_NORMAL,
    THUMB_LARGE,
    USER_PLUGINS,
    USER_CSS,
)

LIB_PATH = os.path.join(USER_PLUGINS, "lib")
if LIB_PATH not in sys.path:
    sys.path.append(LIB_PATH)


# -------------------------------------------------------------------------
#
# Paths to python modules - assumes that the root directory is one level
# above this one, and that the plugins directory is below the root directory.
#
# -------------------------------------------------------------------------
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))

sys.path.insert(0, ROOT_DIR)
git_revision = get_git_revision(ROOT_DIR).replace("\n", "")
if sys.platform == "win32" and git_revision == "":
    git_revision = get_git_revision(os.path.split(ROOT_DIR)[1])
if DEV_VERSION:
    VERSION += git_revision
# VERSION += "-1"

#
# Glade files
#
GLADE_DIR = os.path.join(ROOT_DIR, "gui", "glade")
GLADE_FILE = os.path.join(GLADE_DIR, "gramps.glade")
PERSON_GLADE = os.path.join(GLADE_DIR, "edit_person.glade")
PLUGINS_GLADE = os.path.join(GLADE_DIR, "plugins.glade")
MERGE_GLADE = os.path.join(GLADE_DIR, "mergedata.glade")
RULE_GLADE = os.path.join(GLADE_DIR, "rule.glade")


PLUGINS_DIR = os.path.join(ROOT_DIR, "plugins")

USE_TIPS = False

# -------------------------------------------------------------------------
#
# Paths to data files.
#
# -------------------------------------------------------------------------
_resources = ResourcePath()
DATA_DIR = _resources.data_dir
IMAGE_DIR = _resources.image_dir

TIP_DATA = os.path.join(DATA_DIR, "tips.xml")
PAPERSIZE = os.path.join(DATA_DIR, "papersize.xml")

ICON = os.path.join(IMAGE_DIR, "gramps.png")
LOGO = os.path.join(IMAGE_DIR, "logo.png")
SPLASH = os.path.join(IMAGE_DIR, "splash.jpg")

LICENSE_FILE = os.path.join(_resources.doc_dir, "COPYING")

# -------------------------------------------------------------------------
#
# Gramps environment variables dictionary
#
# -------------------------------------------------------------------------
ENV = {
    "USER_HOME": USER_HOME,
    "USER_CACHE": USER_CACHE,
    "USER_CONFIG": USER_CONFIG,
    "USER_DATA": USER_DATA,
    "VERSION": VERSION,
    "major_version": major_version,
    "VERSION_DIR": VERSION_DIR,
    "USER_DATA_VERSION": USER_DATA_VERSION,
    "THUMB_DIR": THUMB_DIR,
    "THUMB_NORMAL": THUMB_NORMAL,
    "THUMB_LARGE": THUMB_LARGE,
    "USER_PLUGINS": USER_PLUGINS,
    "ROOT_DIR": ROOT_DIR,
    "GLADE_DIR": GLADE_DIR,
    "PLUGINS_DIR": PLUGINS_DIR,
    "DATA_DIR": DATA_DIR,
    "IMAGE_DIR": IMAGE_DIR,
}

# -------------------------------------------------------------------------
#
# Init Localization
#
# -------------------------------------------------------------------------
GRAMPS_LOCALE = GrampsLocale(localedir=_resources.locale_dir)
_ = GRAMPS_LOCALE.translation.sgettext
GTK_GETTEXT_DOMAIN = "gtk30"

# -------------------------------------------------------------------------
#
# About box information
#
# -------------------------------------------------------------------------
COPYRIGHT_MSG = "© 2001-2006 Donald N. Allingham\n" "© 2007-2025 The Gramps Developers"
COMMENTS = _(
    "Gramps is a genealogy program that is both intuitive for hobbyists "
    "and feature-complete for professional genealogists."
)
AUTHORS = [
    "Alexander Roitman",
    "Benny Malengier",
    "Brian Matherly",
    "Donald A. Peterson",
    "Donald N. Allingham",
    "David Hampton",
    "Martin Hawlisch",
    "Nick Hall",
    "Richard Taylor",
    "Tim Waugh",
    "John Ralls",
]

AUTHORS_FILE = os.path.join(DATA_DIR, "authors.xml")

DOCUMENTERS = [
    "Alexander Roitman",
]

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
THUMBSCALE = 96.0
THUMBSCALE_LARGE = 180.0
SIZE_NORMAL = 0
SIZE_LARGE = 1
XMLFILE = "data.gramps"
NO_SURNAME = "(%s)" % _("none", "surname")
NO_GIVEN = "(%s)" % _("none", "given-name")
ARABIC_COMMA = "،"
ARABIC_SEMICOLON = "؛"
DOCGEN_OPTIONS = "Docgen Options"
COLON = _(":")  # Translators: needed for French, ignore otherwise

# -------------------------------------------------------------------------
#
# Options Constants
#
# -------------------------------------------------------------------------

LONGOPTS = [
    "action=",
    "class=",
    "config=",
    "debug=",
    "default=",
    "display=",
    "disable-sound",
    "disable-crash-dialog",
    "enable-sound",
    "espeaker=",
    "export=",
    "force-unlock",
    "format=",
    "gdk-debug=",
    "gdk-no-debug=",
    "gtk-debug=",
    "gtk-no-debug=",
    "gtk-module=",
    "g-fatal-warnings",
    "help",
    "import=",
    "load-modules=",
    "list" "name=",
    "oaf-activate-iid=",
    "oaf-ior-fd=",
    "oaf-private",
    "open=",
    "username=",
    "password=",
    "create=",
    "options=",
    "safe",
    "screen=",
    "show",
    "sm-client-id=",
    "sm-config-prefix=",
    "sm-disable",
    "sync",
    "remove=",
    "usage",
    "version",
    "yes",
    "quiet",
]

SHORTOPTS = "O:U:P:C:i:e:f:a:p:d:c:r:lLthuv?syqSD:"

GRAMPS_UUID = uuid.UUID("516cd010-5a41-470f-99f8-eb22f1098ad6")

# -------------------------------------------------------------------------
#
# Fanchart Constants
#
# -------------------------------------------------------------------------

PIXELS_PER_GENERATION = 50  # size of radius for generation
BORDER_EDGE_WIDTH = 10  # empty white box size at edge to indicate parents
CHILDRING_WIDTH = 12  # width of the children ring inside the person
TRANSLATE_PX = 10  # size of the central circle, used to move the chart
PAD_PX = 4  # padding with edges
PAD_TEXT = 2  # padding for text in boxes

BACKGROUND_SCHEME1 = 0
BACKGROUND_SCHEME2 = 1
BACKGROUND_GENDER = 2
BACKGROUND_WHITE = 3
BACKGROUND_GRAD_GEN = 4
BACKGROUND_GRAD_AGE = 5
BACKGROUND_SINGLE_COLOR = 6
BACKGROUND_GRAD_PERIOD = 7
GENCOLOR = {
    BACKGROUND_SCHEME1: (
        (255, 63, 0),
        (255, 175, 15),
        (255, 223, 87),
        (255, 255, 111),
        (159, 255, 159),
        (111, 215, 255),
        (79, 151, 255),
        (231, 23, 255),
        (231, 23, 121),
        (210, 170, 124),
        (189, 153, 112),
    ),
    BACKGROUND_SCHEME2: (
        (229, 191, 252),
        (191, 191, 252),
        (191, 222, 252),
        (183, 219, 197),
        (206, 246, 209),
    ),
    BACKGROUND_WHITE: (
        (255, 255, 255),
        (255, 255, 255),
    ),
}

MAX_AGE = 100
GRADIENTSCALE = 5

FORM_CIRCLE = 0
FORM_HALFCIRCLE = 1
FORM_QUADRANT = 2

COLLAPSED = 0
NORMAL = 1
EXPANDED = 2

TYPE_BOX_NORMAL = 0
TYPE_BOX_FAMILY = 1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2015       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from copy import deepcopy

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .editreference import RefTab, EditReference
from ..glade import Glade
from ..widgets import (
    MonitoredDate,
    MonitoredEntry,
    MonitoredDataType,
    PrivacyButton,
    MonitoredTagList,
)
from .displaytabs import (
    PlaceRefEmbedList,
    PlaceNameEmbedList,
    LocationEmbedList,
    CitationEmbedList,
    GalleryTab,
    NoteTab,
    WebEmbedList,
    PlaceBackRefList,
)
from gramps.gen.lib import NoteType
from gramps.gen.db import DbTxn
from gramps.gen.errors import ValidationError, WindowActiveError
from gramps.gen.utils.place import conv_lat_lon
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.config import config
from ..dialog import ErrorDialog
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# EditPlaceRef class
#
# -------------------------------------------------------------------------
class EditPlaceRef(EditReference):
    def __init__(self, state, uistate, track, place, place_ref, update):
        EditReference.__init__(self, state, uistate, track, place, place_ref, update)
        self.original = deepcopy(place.serialize())

    def _local_init(self):
        self.top = Glade()
        self.set_window(self.top.toplevel, None, _("Place Reference Editor"))
        self.setup_configs("interface.place-ref", 600, 450)

        self.define_warn_box(self.top.get_object("warning"))
        self.define_expander(self.top.get_object("expander"))
        # self.place_name_label = self.top.get_object('place_name_label')
        # self.place_name_label.set_text(_('Name:', 'place'))

        tblref = self.top.get_object("table64")
        notebook = self.top.get_object("notebook_ref")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.reftab = RefTab(
            self.dbstate, self.uistate, self.track, _("General"), tblref
        )
        tblref = self.top.get_object("table62")
        notebook = self.top.get_object("notebook")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.primtab = RefTab(
            self.dbstate, self.uistate, self.track, _("_General"), tblref
        )

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("cancel"))
        self.ok_button = self.top.get_object("ok")
        self.define_ok_button(self.ok_button, self.save)
        self.define_help_button(self.top.get_object("help"))

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("place-rebuild", self.close)
        self._add_db_signal("place-delete", self.check_for_close)

    def build_menu_names(self, placeref):
        if self.source and self.source.get_handle():
            title = self.source.get_title()
            submenu_label = _("Place: %s") % title
        else:
            submenu_label = _("New Place")
        return (_("Place Reference Editor"), submenu_label)

    def _setup_fields(self):
        self.date_field = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_stat"),
            self.source_ref.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        if not config.get("preferences.place-auto"):
            self.top.get_object("place_title").show()
            self.top.get_object("place_title_label").show()
            self.title = MonitoredEntry(
                self.top.get_object("place_title"),
                self.source.set_title,
                self.source.get_title,
                self.db.readonly,
            )

        self.name = MonitoredEntry(
            self.top.get_object("name_entry"),
            self.source.get_name().set_value,
            self.source.get_name().get_value,
            self.db.readonly,
            changed=self.name_changed,
        )

        edit_button = self.top.get_object("name_button")
        edit_button.connect("clicked", self.edit_place_name)

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.source.set_gramps_id,
            self.source.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.source.set_tag_list,
            self.source.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.privacy = PrivacyButton(
            self.top.get_object("private"), self.source, self.db.readonly
        )

        custom_place_types = sorted(self.db.get_place_types(), key=lambda s: s.lower())
        self.place_type = MonitoredDataType(
            self.top.get_object("place_type"),
            self.source.set_type,
            self.source.get_type,
            self.db.readonly,
            custom_place_types,
        )

        self.code = MonitoredEntry(
            self.top.get_object("code_entry"),
            self.source.set_code,
            self.source.get_code,
            self.db.readonly,
        )

        entry = self.top.get_object("lon_entry")
        entry.set_ltr_mode()
        self.longitude = MonitoredEntry(
            entry,
            self.source.set_longitude,
            self.source.get_longitude,
            self.db.readonly,
        )
        self.longitude.connect("validate", self._validate_coordinate, "lon")
        # force validation now with initial entry
        entry.validate(force=True)

        entry = self.top.get_object("lat_entry")
        entry.set_ltr_mode()
        self.latitude = MonitoredEntry(
            entry, self.source.set_latitude, self.source.get_latitude, self.db.readonly
        )
        self.latitude.connect("validate", self._validate_coordinate, "lat")
        # force validation now with initial entry
        entry.validate(force=True)

        entry = self.top.get_object("latlon_entry")
        entry.set_ltr_mode()
        self.latlon = MonitoredEntry(
            entry, self.set_latlongitude, self.get_latlongitude, self.db.readonly
        )

    def set_latlongitude(self, value):
        try:
            # Bug 12349, 12374
            parts = value.split(", ")
            if len(parts) == 2:
                latitude = parts[0].strip().replace(",", ".")
                longitude = parts[1].strip().replace(",", ".")
            else:
                latitude, longitude = value.split(",")

            self.longitude.set_text(longitude)
            self.latitude.set_text(latitude)
            self.top.get_object("lat_entry").validate(force=True)
            self.top.get_object("lon_entry").validate(force=True)
            self.source.set_latitude(self.latitude.get_value())
            self.source.set_longitude(self.longitude.get_value())
        except:
            pass

    def get_latlongitude(self):
        return ""

    def _validate_coordinate(self, widget, text, typedeg):
        if (typedeg == "lat") and not conv_lat_lon(text, "0", "ISO-D"):
            return ValidationError(
                # Translators: translate the "S" too (and the "or" of course)
                _(
                    "Invalid latitude\n(syntax: "
                    "18\u00b09'48.21\"S, -18.2412 or -18:9:48.21)"
                )
            )
        elif (typedeg == "lon") and not conv_lat_lon("0", text, "ISO-D"):
            return ValidationError(
                # Translators: translate the "E" too (and the "or" of course)
                _(
                    "Invalid longitude\n(syntax: "
                    "18\u00b09'48.21\"E, -18.2412 or -18:9:48.21)"
                )
            )

    def update_title(self):
        new_title = place_displayer.display(self.db, self.source)
        self.top.get_object("preview_title").set_text(new_title)

    def name_changed(self, obj):
        self.update_title()

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.

        """
        notebook = self.top.get_object("notebook")
        notebook_ref = self.top.get_object("notebook_ref")

        self._add_tab(notebook, self.primtab)
        self._add_tab(notebook_ref, self.reftab)
        self.track_ref_for_deletion("primtab")
        self.track_ref_for_deletion("reftab")

        self.placeref_list = PlaceRefEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_placeref_list(),
            "placeref_editor_placerefs",
            self.source.handle,
            self.update_title,
        )
        self._add_tab(notebook, self.placeref_list)
        self.track_ref_for_deletion("placeref_list")

        self.alt_name_list = PlaceNameEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.alt_names,
            "placeref_editor_altnames",
        )
        self._add_tab(notebook, self.alt_name_list)
        self.track_ref_for_deletion("alt_name_list")

        if len(self.source.alt_loc) > 0:
            self.loc_list = LocationEmbedList(
                self.dbstate,
                self.uistate,
                self.track,
                self.source.alt_loc,
                "placeref_editor_locations",
            )
            self._add_tab(notebook, self.loc_list)
            self.track_ref_for_deletion("loc_list")

        self.citation_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_citation_list(),
            "placeref_editor_citations",
        )
        self._add_tab(notebook, self.citation_list)
        self.track_ref_for_deletion("citation_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_note_list(),
            "placeref_editor_notes",
            notetype=NoteType.PLACE,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.web_list = WebEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_url_list(),
            "placeref_editor_internet",
        )
        self._add_tab(notebook, self.web_list)
        self.track_ref_for_deletion("web_list")

        self.backref_list = PlaceBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.source.handle),
            "placeref_editor_references",
        )
        self.backref_tab = self._add_tab(notebook, self.backref_list)
        self.track_ref_for_deletion("backref_list")
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook)

    def edit_place_name(self, obj):
        try:
            from . import EditPlaceName

            EditPlaceName(
                self.dbstate,
                self.uistate,
                self.track,
                self.source.get_name(),
                self.edit_callback,
            )
        except WindowActiveError:
            return

    def edit_callback(self, obj):
        value = self.source.get_name().get_value()
        self.top.get_object("name_entry").set_text(value)

    def save(self, *obj):
        self.ok_button.set_sensitive(False)

        if self.source.get_name().get_value().strip() == "":
            msg1 = _("Cannot save place. Name not entered.")
            msg2 = _("You must enter a name before saving.")
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        if self.source.handle:
            # only commit if it has changed
            if self.source.serialize() != self.original:
                with DbTxn(_("Modify Place"), self.db) as trans:
                    self.db.commit_place(self.source, trans)
        else:
            if self.check_for_duplicate_id("Place"):
                return
            with DbTxn(_("Add Place"), self.db) as trans:
                self.db.add_place(self.source, trans)
            self.source_ref.ref = self.source.handle

        if self.update:
            self.update(self.source_ref, self.source)

        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""The core library of Gramps objects
"""
# Dates
from .date import Date, DateError, Span

# Secondary objects
from .secondaryobj import SecondaryObject
from .address import Address
from .location import Location
from .attribute import Attribute
from .srcattribute import SrcAttribute
from .eventref import EventRef
from .ldsord import LdsOrd
from .mediaref import MediaRef
from .name import Name
from .placename import PlaceName
from .placeref import PlaceRef
from .reporef import RepoRef
from .surname import Surname
from .url import Url
from .childref import ChildRef

# Primary objects
from .primaryobj import PrimaryObject
from .person import Person
from .personref import PersonRef
from .family import Family
from .event import Event
from .place import Place
from .src import Source
from .media import Media
from .repo import Repository
from .note import Note
from .citation import Citation

# Table objects
from .tag import Tag

# These are actually metadata
from .genderstats import GenderStats
from .researcher import Researcher

# Type classes
from .grampstype import GrampsType
from .nametype import NameType
from .attrtype import AttributeType
from .srcattrtype import SrcAttributeType
from .urltype import UrlType
from .childreftype import ChildRefType
from .repotype import RepositoryType
from .eventtype import EventType
from .familyreltype import FamilyRelType
from .srcmediatype import SourceMediaType
from .eventroletype import EventRoleType
from .markertype import MarkerType
from .nameorigintype import NameOriginType
from .notetype import NoteType
from .styledtexttagtype import StyledTextTagType
from .placetype import PlaceType

# Text
from .styledtexttag import StyledTextTag
from .styledtext import StyledText



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Database Processing/Rename Event Types"""

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gui.utils import ProgressMeter
from gramps.gui.managedwindow import ManagedWindow
from gramps.gen.lib import EventType
from gramps.gen.db import DbTxn
from gramps.gui.plug import tool
from gramps.gui.glade import Glade


# -------------------------------------------------------------------------
#
# ChangeTypes class
#
# -------------------------------------------------------------------------
class ChangeTypes(tool.BatchTool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        self.user = user
        self.dbstate = dbstate

        tool.BatchTool.__init__(self, dbstate, user, options_class, name)
        if self.fail:
            return

        if uistate:
            self.title = _("Change Event Types")
            ManagedWindow.__init__(self, uistate, [], self.__class__)
            self.init_gui()
        else:
            self.run_tool()

    def init_gui(self):
        # Draw dialog and make it handle everything

        self.glade = Glade()

        self.auto1 = self.glade.get_object("original")
        self.auto2 = self.glade.get_object("new")

        # Need to display localized event names
        etype = EventType()
        custom_events = self.dbstate.db.get_event_types()
        event_names = sorted(
            etype.get_standard_names() + custom_events, key=glocale.sort_key
        )

        self.fill_combo(self.auto1, event_names)
        self.fill_combo(self.auto2, event_names)

        etype.set_from_xml_str(self.options.handler.options_dict["fromtype"])
        self.auto1.get_child().set_text(str(etype))

        etype.set_from_xml_str(self.options.handler.options_dict["totype"])
        self.auto2.get_child().set_text(str(etype))

        window = self.glade.toplevel
        self.set_window(window, self.glade.get_object("title"), self.title)
        self.setup_configs("interface.changetypes", 640, 260)

        self.glade.connect_signals(
            {
                "on_close_clicked": self.close,
                "on_apply_clicked": self.on_apply_clicked,
                "on_delete_event": self.close,
            }
        )

        self.show()

    def build_menu_names(self, obj):
        return (self.title, None)

    def run_tool(self, parent_window=None):
        # Run tool and return results
        # These are English names, no conversion needed
        fromtype = self.options.handler.options_dict["fromtype"]
        totype = self.options.handler.options_dict["totype"]

        modified = 0

        with DbTxn(_("Change types"), self.db, batch=True) as self.trans:
            self.db.disable_signals()
            with self.user.progress(
                _("Analyzing Events"), "", self.db.get_number_of_events()
            ) as step:
                for event_handle in self.db.get_event_handles():
                    event = self.db.get_event_from_handle(event_handle)
                    if event.get_type().xml_str() == fromtype:
                        event.type.set_from_xml_str(totype)
                        modified += 1
                        self.db.commit_event(event, self.trans)
                        step()
        self.db.enable_signals()
        self.db.request_rebuild()

        if modified == 0:
            msg = _("No event record was modified.")
        else:
            # Translators: leave all/any {...} untranslated
            msg = ngettext(
                "{number_of} event record was modified.",
                "{number_of} event records were modified.",
                modified,
            ).format(number_of=modified)

        self.user.info(_("Change types"), msg, parent=parent_window)
        return (bool(modified), msg)

    def on_apply_clicked(self, obj):
        # Need to store English names for later comparison
        the_type = EventType()

        the_type.set(self.auto1.get_child().get_text())
        self.options.handler.options_dict["fromtype"] = the_type.xml_str()

        the_type.set(self.auto2.get_child().get_text())
        self.options.handler.options_dict["totype"] = the_type.xml_str()

        self.run_tool(self.parent_window)

        # Save options
        self.options.handler.save_options()

        self.close()

    def fill_combo(self, combo, data_list):
        """
        Fill a combo box with completion data
        """
        store = Gtk.ListStore(GObject.TYPE_STRING)

        for data in data_list:
            if data:
                store.append(row=[data])

        combo.set_popup_fixed_width(False)
        combo.set_wrap_width(4)
        combo.set_model(store)
        combo.set_entry_text_column(0)
        completion = Gtk.EntryCompletion()
        completion.set_model(store)
        completion.set_minimum_key_length(1)
        completion.set_text_column(0)
        combo.get_child().set_completion(completion)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class ChangeTypesOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "fromtype": "",
            "totype": "",
        }
        self.options_help = {
            "fromtype": ("=str", "Type of events to replace", "Event type string"),
            "totype": ("=str", "New type replacing the old one", "Event type string"),
        }



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
# Copyright (C) 2018       Paul D.Smith
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


class LayoutTree:
    """
    Narrative Web Page generator.

    Classe:
        BuchheimTree - A tree suitable for passing to the Bushheim layout
        algorithm
    """

    def __init__(self, handle, *children):
        self.handle = handle
        self.children = []
        if isinstance(children, list):
            children = list(children)
        for parent in children:
            if parent is not None:
                self.children.append(parent)

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return self.children[key]
        if isinstance(key, str):
            for child in self.children:
                if child.node == key:
                    return child
        assert "Key not found"
        return None

    def __iter__(self):
        return self.children.__iter__()

    def __len__(self):
        return len(self.children)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""class for generating dialogs for graphviz-based reports """

# -------------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------------
from ._graphreportdialog import GraphReportDialog, BaseFormatComboBox
from gramps.gen.plug.report import CATEGORY_GRAPHVIZ
import gramps.gen.plug.docgen.graphdoc as graphdoc


# -----------------------------------------------------------------------
#
# GraphvizReportDialog
#
# -----------------------------------------------------------------------
class GraphvizReportDialog(GraphReportDialog):
    def make_doc_menu(self):
        """
        Build a menu of document types that are appropriate for
        this graph report.
        """
        self.format_menu = GraphvizFormatComboBox()

    def get_category(self):
        """
        Return the report category.
        """
        return CATEGORY_GRAPHVIZ

    def get_options(self):
        """
        Return the graph options.
        """
        return graphdoc.GVOptions()

    def doc_type_changed(self, obj):
        """
        This routine is called when the user selects a new file
        format for the report.  It adjusts the various dialog sections
        to reflect the appropriate values for the currently selected
        file format.  For example, a HTML document doesn't need any
        paper size/orientation options, but it does need a template
        file.  Those changes are made here.
        """
        GraphReportDialog.doc_type_changed(self, obj)

        output_format_str = obj.get_clname()
        if output_format_str in ["gvpdf", "gspdf", "ps"]:
            # Always use 72 DPI for PostScript and PDF files.
            self._goptions.dpi.set_value(72)
            self._goptions.dpi.set_available(False)
        else:
            self._goptions.dpi.set_available(True)

        if output_format_str in ["gspdf", "dot"]:
            # Multiple pages only valid for dot and PDF via GhostsScript.
            self._goptions.h_pages.set_available(True)
            self._goptions.v_pages.set_available(True)
        else:
            self._goptions.h_pages.set_value(1)
            self._goptions.v_pages.set_value(1)
            self._goptions.h_pages.set_available(False)
            self._goptions.v_pages.set_available(False)


# -------------------------------------------------------------------------------
#
# GraphvizFormatComboBox
#
# -------------------------------------------------------------------------------
class GraphvizFormatComboBox(BaseFormatComboBox):
    FORMATS = graphdoc.FORMATS



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
CitationBase class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".citation")


# -------------------------------------------------------------------------
#
# CitationBase class
#
# -------------------------------------------------------------------------
class CitationBase:
    """
    Base class for storing citations.

    Starting in 3.4, the objects may have multiple citations.
    Internally, this class maintains a list of Citation handles,
    as a citation_list attribute of the CitationBase object.
    This class is analogous to the notebase class.
    Both these have no attributes of their own; in this respect, they differ
    from classes like MediaRef, which does have attributes (in that case,
    privacy, sources, notes and attributes).

    This class, together with the Citation class, replaces the old SourceRef
    class. I.e. SourceRef = CitationBase + Citation
    """

    def __init__(self, source=None):
        """
        Create a new CitationBase, copying from source if not None.

        :param source: Object used to initialize the new object
        :type source: CitationBase
        """
        self.citation_list = list(source.citation_list) if source else []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return self.citation_list

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.citation_list = list(data)

    def add_citation(self, handle):
        """
        Add the :class:`~.citation.Citation` handle to the list of citation
        handles.

        :param handle: :class:`~.citation.Citation` handle to add the list of
                       citations
        :type handle: str

        :returns: True if handle was added, False if it already was in the list
        :rtype: bool
        """
        if handle in self.citation_list:
            return False
        else:
            self.citation_list.append(handle)
            return True

    def remove_citation_references(self, citation_handle_list):
        """
        Remove the specified handles from the list of citation handles, and all
        secondary child objects.

        :param citation_handle_list: The list of citation handles to be removed
        :type handle: list
        """
        LOG.debug(
            "enter remove_citation handle: %s self: %s citation_list: %s",
            citation_handle_list,
            self,
            self.citation_list,
        )
        for handle in citation_handle_list:
            if handle in self.citation_list:
                LOG.debug(
                    "remove handle %s from citation_list %s", handle, self.citation_list
                )
                self.citation_list.remove(handle)
        LOG.debug("get_citation_child_list %s", self.get_citation_child_list())
        for item in self.get_citation_child_list():
            item.remove_citation_references(citation_handle_list)

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        All methods which inherit from CitationBase and have other child objects
        with citations, should return here a list of child objects which are
        CitationBase

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return []

    def get_citation_list(self):
        """
        Return the list of :class:`~.citation.Citation` handles associated with
        the object.

        :returns: The list of :class:`~.citation.Citation` handles
        :rtype: list
        """
        return self.citation_list

    def has_citations(self):
        """
        Return True if citations exist for this object.

        :returns: Returns whether citations exist for this object.
        :rtype: bool
        """
        return bool(self.get_citation_list())

    def get_all_citation_lists(self):
        """
        Return the list of :class:`~.citation.Citation` handles associated with
        the object or with child objects.

        :returns: The list of :class:`~.citation.Citation` handles
        :rtype: list
        """
        all_citations = self.citation_list

        for item in self.get_citation_child_list():
            all_citations += item.get_citation_list()
            for subitem in item.get_citation_child_list():
                all_citations += subitem.get_citation_list()
        return all_citations

    def has_citation_reference(self, citation_handle):
        """
        Return True if the object or any of its child objects has reference
        to this citation handle.

        :param citation_handle: The citation handle to be checked.
        :type citation_handle: str
        :returns: Returns whether the object or any of its child objects has
                  reference to this citation handle.
        :rtype: bool
        """
        for citation_ref in self.citation_list:
            if citation_ref == citation_handle:
                return True

        LOG.debug("citation child list %s", self.get_citation_child_list())
        for item in self.get_citation_child_list():
            if item.has_citation_reference(citation_handle):
                return True

        return False

    def set_citation_list(self, citation_list):
        """
        Assign the passed list to be object's list of
        :class:`~.citation.Citation` handles.

        :param citation_list: List of :class:`~.citation.Citation` handles to
                              be set on the object
        :type citation_list: list
        """
        self.citation_list = citation_list

    def _merge_citation_list(self, acquisition):
        """
        Merge the list of citations from acquisition with our own.

        :param acquisition: The citation list of this object will be merged
                            with the current citation list.
        :type acquisition: CitationBase
        """
        for addendum in acquisition.citation_list:
            self.add_citation(addendum)

    def get_referenced_citation_handles(self):
        """
        Return the list of (classname, handle) tuples for all referenced
        citations.

        This method should be used to get the :class:`~.citation.Citation`
        portion of the list by objects that store citation lists.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return [("Citation", handle) for handle in self.citation_list]

    def replace_citation_references(self, old_handle, new_handle):
        """
        Replace references to citation handles in the list of this object and
        all child objects and merge equivalent entries.

        :param old_handle: The citation handle to be replaced.
        :type old_handle: str
        :param new_handle: The citation handle to replace the old one with.
        :type new_handle: str
        """
        refs_list = self.citation_list[:]
        new_ref = None
        if new_handle in self.citation_list:
            new_ref = new_handle
        n_replace = refs_list.count(old_handle)
        for ix_replace in range(n_replace):
            idx = refs_list.index(old_handle)
            if new_ref:
                self.citation_list.pop(idx)
                refs_list.pop(idx)
            else:
                self.citation_list[idx] = new_handle

        for item in self.get_citation_child_list():
            item.replace_citation_references(old_handle, new_handle)


class IndirectCitationBase:
    """
    Citation management logic for objects that don't have citations
    for the primary objects, but only for the child (secondary) ones.

    The derived class must implement the
    :meth:`~CitationBase.get_citation_child_list` method to return the list of
    child secondary objects that may refer citations.

    .. note:: for most objects, this functionality is inherited from
              :class:`CitationBase`, which checks both the object and the child
              objects.
    """

    def has_citation_reference(self, citation_handle):
        """
        Return True if any of the child objects has reference to this citation
        handle.

        :param citation_handle: The citation handle to be checked.
        :type citation_handle: str
        :returns: Returns whether any of it's child objects has reference to
                  this citation handle.
        :rtype: bool
        """
        for item in self.get_citation_child_list():
            if item.has_citation_reference(citation_handle):
                return True

        return False

    def replace_citation_references(self, old_handle, new_handle):
        """
        Replace references to citation handles in all child objects and merge
        equivalent entries.

        :param old_handle: The citation handle to be replaced.
        :type old_handle: str
        :param new_handle: The citation handle to replace the old one with.
        :type new_handle: str
        """
        for item in self.get_citation_child_list():
            item.replace_citation_references(old_handle, new_handle)

    def remove_citation_references(self, citation_handle_list):
        """
        Remove references to all citation handles in the list in all child
        objects.

        :param citation_handle_list: The list of citation handles to be removed.
        :type citation_handle_list: list
        """
        for item in self.get_citation_child_list():
            item.remove_citation_references(citation_handle_list)

    def get_citation_list(self):
        """
        Return the list of :class:`~gen.lib.citation.Citation` handles
        associated with the object. For an IndirectCitationBase this is always
        the empty list
        :returns: The list of :class:`~gen.lib.citation.Citation` handles
        :rtype: list
        """
        return []

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        All methods which inherit from CitationBase and have other child objects
        with citations, should return here a list of child objects which are
        CitationBase

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return []



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..managedwindow import ManagedWindow
from ..glade import Glade
from ..listmodel import ListModel
from gramps.gen.errors import ValidationError
from gramps.gen.display.place import displayer as _pd
from gramps.gen.display.place import PlaceFormat
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# EditPlaceFormat
#
# -------------------------------------------------------------------------
class EditPlaceFormat(ManagedWindow):
    def __init__(self, uistate, dbstate, track, callback):
        self.title = _("Place Format Editor")
        ManagedWindow.__init__(self, uistate, track, EditPlaceFormat)
        self.callback = callback
        self.top = Glade()
        self.set_window(self.top.toplevel, None, self.title, None)
        self.setup_configs("interface.editplaceformat", 600, 400)
        self.top.get_object("add").connect("clicked", self.__add)
        self.top.get_object("remove").connect("clicked", self.__remove)
        self.top.get_object("name").connect("changed", self.__name_changed)
        self.top.get_object("levels").connect("validate", self._validate)
        self.window.connect("response", self.__close)
        self.model = None
        self.formats = _pd.get_formats()
        self.current_format = None
        self.__populate_format_list()
        self.show()

    def build_menu_names(self, obj):
        return (self.title, None)

    def __populate_format_list(self):
        flist = self.top.get_object("format_list")
        self.model = ListModel(
            flist, [(_("Format"), -1, 100)], select_func=self.__format_changed
        )
        for fmt in self.formats:
            self.model.add([fmt.name])
        self.model.select_row(0)

    def __format_changed(self, selection):
        if self.current_format is not None:
            fmt = self.formats[self.current_format]
            self.__save_format(fmt)
        row = self.model.get_selected_row()
        if row != -1:
            fmt = self.formats[row]
            self.__load_format(fmt)
            self.current_format = row
        if row == 0:
            self.top.get_object("remove").set_sensitive(False)
            self.top.get_object("name").set_sensitive(False)
            self.top.get_object("levels").set_sensitive(False)
            self.top.get_object("street").set_sensitive(False)
            self.top.get_object("language").set_sensitive(False)
            self.top.get_object("reverse").set_sensitive(False)
        else:
            self.top.get_object("remove").set_sensitive(True)
            self.top.get_object("name").set_sensitive(True)
            self.top.get_object("levels").set_sensitive(True)
            self.top.get_object("street").set_sensitive(True)
            self.top.get_object("language").set_sensitive(True)
            self.top.get_object("reverse").set_sensitive(True)
        self.top.get_object("levels").validate(force=True)

    def __name_changed(self, entry):
        store, iter_ = self.model.get_selected()
        self.model.set(iter_, [entry.get_text()])

    def _validate(self, widget, text):
        for level in text.split(","):
            parts = level.split(":")
            if len(parts) < 1:
                return ValidationError("Empty level")
            if len(parts) > 2:
                return ValidationError("Invalid slice")
            for part in parts:
                integer_str = part.replace("p", "")
                if integer_str != "":
                    try:
                        integer = int(integer_str)
                    except ValueError:
                        return ValidationError("Invalid format string")

    def __load_format(self, fmt):
        self.top.get_object("name").set_text(fmt.name)
        self.top.get_object("levels").set_text(fmt.levels)
        self.top.get_object("street").set_active(fmt.street)
        self.top.get_object("language").set_text(fmt.language)
        self.top.get_object("reverse").set_active(fmt.reverse)

    def __save_format(self, fmt):
        fmt.name = self.top.get_object("name").get_text()
        fmt.levels = self.top.get_object("levels").get_text()
        fmt.street = self.top.get_object("street").get_active()
        fmt.language = self.top.get_object("language").get_text()
        fmt.reverse = self.top.get_object("reverse").get_active()

    def __add(self, button):
        name = _("New")
        self.formats.append(PlaceFormat(name, ":", "", 0, False))
        self.model.add([name])
        self.model.select_row(len(self.formats) - 1)

    def __remove(self, button):
        store, iter_ = self.model.get_selected()
        if iter_:
            self.current_format = None
            del self.formats[self.model.get_selected_row()]
            self.model.remove(iter_)
        if self.model.get_selected_row() == -1:
            self.model.select_row(len(self.formats) - 1)

    def __close(self, *obj):
        row = self.model.get_selected_row()
        fmt = self.formats[self.current_format]
        self.__save_format(fmt)
        _pd.save_formats()
        self.callback()
        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons, Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
CitationTreeModel classes for Gramps.
"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")
LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.db import get_source_referents
from .treebasemodel import TreeBaseModel
from .citationbasemodel import CitationBaseModel


# -------------------------------------------------------------------------
#
# CitationModel
#
# -------------------------------------------------------------------------
class CitationTreeModel(CitationBaseModel, TreeBaseModel):
    """
    Hierarchical citation model.
    """

    def __init__(
        self,
        db,
        uistate,
        scol=0,
        order=Gtk.SortType.ASCENDING,
        search=None,
        skip=set(),
        sort_map=None,
    ):
        self.db = db
        self.number_items = self.db.get_number_of_sources
        self.map = self.db.get_raw_source_data
        self.gen_cursor = self.db.get_source_cursor
        # The items here must correspond, in order, with data in
        # CitationTreeView, and with the items in the secondary fmap, fmap2
        self.fmap = [
            self.source_src_title,  # COL_TITLE_PAGE (both Source & Citation)
            self.source_src_id,  # COL_ID         (both Source & Citation)
            None,  # COL_DATE       (not for Source)
            None,  # COL_CONFIDENCE (not for Source)
            self.source_src_private,  # COL_PRIV       (both Source & Citation)
            self.source_src_tags,  # COL_TAGS       (both Source & Citation)
            self.source_src_chan,  # COL_CHAN       (both Source & Citation)
            self.source_src_auth,  # COL_SRC_AUTH   (Source only)
            self.source_src_abbr,  # COL_SRC_ABBR   (Source only)
            self.source_src_pinfo,  # COL_SRC_PINFO  (Source only)
            self.source_src_tag_color,
        ]
        self.smap = [
            self.source_src_title,
            self.source_src_id,
            self.dummy_sort_key,
            self.dummy_sort_key,
            self.source_src_private,
            self.source_src_tags,
            self.source_sort2_change,
            self.source_src_auth,
            self.source_src_abbr,
            self.source_src_pinfo,
            self.source_src_tag_color,
        ]

        TreeBaseModel.__init__(
            self,
            self.db,
            uistate,
            scol=scol,
            order=order,
            search=search,
            skip=skip,
            sort_map=sort_map,
            nrgroups=1,
            group_can_have_handle=True,
            has_secondary=True,
        )

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        self.db = None
        self.gen_cursor = None
        self.map = None
        self.fmap = None
        self.smap = None
        self.number_items = None
        self.gen_cursor2 = None
        self.map2 = None
        self.fmap2 = None
        self.smap2 = None
        self.number_items2 = None
        TreeBaseModel.destroy(self)

    def _set_base_data(self):
        """See TreeBaseModel, for citations, most have been set in init of
        CitationBaseModel
        """
        self.number_items2 = self.db.get_number_of_citations
        self.map2 = self.db.get_raw_citation_data
        self.gen_cursor2 = self.db.get_citation_cursor
        self.fmap2 = [
            self.citation_page,
            self.citation_id,
            self.citation_date,
            self.citation_confidence,
            self.citation_private,
            self.citation_tags,
            self.citation_change,
            None,
            None,
            None,
            self.citation_tag_color,
        ]
        self.smap2 = [
            self.citation_page,
            self.citation_id,
            self.citation_sort_date,
            self.citation_sort_confidence,
            self.citation_private,
            self.citation_tags,
            self.citation_sort_change,
            self.dummy_sort_key,
            self.dummy_sort_key,
            self.dummy_sort_key,
            self.citation_tag_color,
        ]

    def color_column(self):
        """
        Return the color column.
        """
        return 10

    def get_tree_levels(self):
        """
        Return the headings of the levels in the hierarchy.
        """
        return [_("Source"), _("Citation")]

    def add_row(self, handle, data):
        """
        Add source nodes to the node map.

        handle      The handle of the gramps object.
        data        The object data.
        """
        sort_key = self.sort_func(data)
        self.add_node(None, handle, sort_key, handle)

    def add_row2(self, handle, data):
        """
        Add citation nodes to the node map.

        handle      The handle of the gramps object.
        data        The object data.
        """
        sort_key = self.sort_func2(data)
        # If the source for this citation already exists (in the tree model) we
        # add the citation as a child of the source. Otherwise we add the source
        # first (because citations don't have any meaning without the associated
        # source)
        if self._get_node(data[5]):
            #             parent   child   sortkey   handle
            self.add_node(data[5], handle, sort_key, handle, secondary=True)
        else:
            # add the source node first
            source_sort_key = self.sort_func(self.map(data[5]))
            #            parent child    sortkey          handle
            self.add_node(None, data[5], source_sort_key, data[5])

            #            parent    child   sortkey   handle
            self.add_node(data[5], handle, sort_key, handle, secondary=True)

    def on_get_n_columns(self):
        return len(self.fmap) + 1

    def column_header(self, node):
        """
        Return a column heading.  This is called for nodes with no associated
        Gramps handle.
        """
        return node.name
